@@ -31,6 +31,7 @@
 #include <language/duchain/types/functiontype.h>
 #include <language/duchain/types/integraltype.h>
 #include <language/duchain/types/structuretype.h>
+#include <language/duchain/types/unsuretype.h>
 
 #include "duchaindebug.h"
 
@@ -43,7 +44,7 @@ namespace Php
 
 ExpressionVisitor::ExpressionVisitor(EditorIntegrator* editor)
         : m_editor(editor), m_createProblems(false),
-          m_offset(CursorInRevision::invalid()), m_currentContext(0),
+          m_offset(CursorInRevision::invalid()), m_currentContext(nullptr),
           m_isAssignmentExpressionEqual(false),
           m_inDefine(false)
 {
@@ -189,19 +190,35 @@ void ExpressionVisitor::visitVariable(VariableAst* node)
     }
 }
 
+void ExpressionVisitor::visitVarExpression(VarExpressionAst *node)
+{
+    DefaultVisitor::visitVarExpression(node);
+    if (node->isGenerator != -1) {
+        DeclarationPointer generatorDecl = findDeclarationImport(ClassDeclarationType, QualifiedIdentifier("generator"));
+
+        if (generatorDecl) {
+            m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeMixed)));
+            if (hasCurrentClosureReturnType()) {
+                FunctionType::Ptr closureType = currentClosureReturnType().cast<FunctionType>();
+                closureType->setReturnType(generatorDecl->abstractType());
+            }
+        }
+    }
+}
+
 void ExpressionVisitor::visitVarExpressionNewObject(VarExpressionNewObjectAst *node)
 {
     DefaultVisitor::visitVarExpressionNewObject(node);
-    if (node->className->staticIdentifier != -1) {
+    if (node->classNameReference->className && node->classNameReference->className->staticIdentifier != -1) {
         static const QualifiedIdentifier id(QStringLiteral("static"));
         DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
-        usingDeclaration(node->className, dec);
+        usingDeclaration(node->classNameReference->className, dec);
         m_result.setDeclaration(dec);
-    } else if (node->className->identifier) {
-        const QualifiedIdentifier id = identifierForNamespace(node->className->identifier, m_editor);
+    } else if (node->classNameReference->className && node->classNameReference->className->identifier) {
+        const QualifiedIdentifier id = identifierForNamespace(node->classNameReference->className->identifier, m_editor);
         DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
-        usingDeclaration(node->className->identifier->namespaceNameSequence->back()->element, dec);
-        buildNamespaceUses(node->className->identifier, id);
+        usingDeclaration(node->classNameReference->className->identifier->namespaceNameSequence->back()->element, dec);
+        buildNamespaceUses(node->classNameReference->className->identifier, id);
         m_result.setDeclaration(dec);
     }
 }
@@ -214,45 +231,46 @@ void ExpressionVisitor::visitVarExpressionArray(VarExpressionArrayAst *node)
 
 void ExpressionVisitor::visitClosure(ClosureAst* node)
 {
-    FunctionType* closureType = new FunctionType;
-    m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeVoid)));
+    auto* closureType = new FunctionType;
+    closureType->setReturnType(AbstractType::Ptr(new IntegralType(IntegralType::TypeVoid)));
+    openClosureReturnType(AbstractType::Ptr(closureType));
     if (node->functionBody) {
         visitInnerStatementList(node->functionBody);
     }
-    closureType->setReturnType(m_result.type());
+    if (node->returnType && node->returnType->typehint && isClassTypehint(node->returnType->typehint, m_editor)) {
+        NamespacedIdentifierAst* objectType = node->returnType->typehint->genericType;
+        QualifiedIdentifier id = identifierForNamespace(objectType, m_editor);
+        DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
+
+        usingDeclaration(objectType->namespaceNameSequence->back()->element, dec);
+        buildNamespaceUses(objectType, id);
+    }
+
+    //Override found type with return typehint or phpdoc return typehint
+    AbstractType::Ptr type = returnType(node->returnType, {}, m_editor, m_currentContext);
+
+    if (type) {
+        closureType->setReturnType(type);
+    }
 
     if (node->parameters->parametersSequence) {
         const KDevPG::ListNode< ParameterAst* >* it = node->parameters->parametersSequence->front();
         forever {
-            AbstractType::Ptr type;
-            if (it->element->parameterType) {
-                //don't use openTypeFromName as it uses cursor for findDeclarations
-                DeclarationPointer decl = findDeclarationImport(ClassDeclarationType,
-                                            identifierForNamespace(it->element->parameterType, m_editor));
-                if (decl) {
-                    type = decl->abstractType();
-                }
-            } else if (it->element->arrayType != -1) {
-                type = AbstractType::Ptr(new IntegralType(IntegralType::TypeArray));
-            } else if (it->element->callableType != -1) {
-                type = AbstractType::Ptr(new IntegralType(IntegralType::TypeMixed));
-            } else if (it->element->defaultValue) {
-                ExpressionVisitor v(m_editor);
-                it->element->defaultValue->ducontext = m_currentContext;
-                v.visitNode(it->element->defaultValue);
-                type = v.result().type();
-            }
-            if (!type) {
-                type = AbstractType::Ptr(new IntegralType(IntegralType::TypeMixed));
-            }
-
-            if ( it->element->isRef != -1 ) {
-                ReferenceType::Ptr p( new ReferenceType() );
-                p->setBaseType( type );
-
-                type = p.cast<AbstractType>();
-            }
+            AbstractType::Ptr type = parameterType(it->element, {}, m_editor, m_currentContext);
             closureType->addArgument(type);
+
+            if (it->element->parameterType && it->element->parameterType->typehint
+                    && isClassTypehint(it->element->parameterType->typehint, m_editor)) {
+                NamespacedIdentifierAst* objectType = it->element->parameterType->typehint->genericType;
+                QualifiedIdentifier id = identifierForNamespace(objectType, m_editor);
+                DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
+
+                usingDeclaration(objectType->namespaceNameSequence->back()->element, dec);
+                buildNamespaceUses(objectType, id);
+            }
+            if (it->element->defaultValue) {
+                visitExpr(it->element->defaultValue);
+            }
             if ( it->hasNext() ) {
                 it = it->next;
             } else {
@@ -282,6 +300,7 @@ void ExpressionVisitor::visitClosure(ClosureAst* node)
     }
 
     m_result.setType(AbstractType::Ptr(closureType));
+    closeClosureReturnType();
 }
 
 void ExpressionVisitor::visitFunctionCallParameterList( FunctionCallParameterListAst* node )
@@ -352,8 +371,12 @@ void ExpressionVisitor::visitFunctionCall(FunctionCallAst* node)
             m_result.setDeclaration(dec);
         } else {
             //global function call foo();
-            const QualifiedIdentifier id = identifierForNamespace(node->stringFunctionNameOrClass, m_editor);
+            QualifiedIdentifier id = identifierForNamespace(node->stringFunctionNameOrClass, m_editor);
             DeclarationPointer dec = findDeclarationImport(FunctionDeclarationType, id);
+            if (!dec) {
+                id.setExplicitlyGlobal(true);
+                dec = findDeclarationImport(FunctionDeclarationType, id);
+            }
             ifDebug(qCDebug(DUCHAIN) << "function call of" << (dec ? dec->toString() : QString("function not found"));)
             m_result.setDeclaration(dec);
             usingDeclaration(node->stringFunctionNameOrClass->namespaceNameSequence->back()->element, dec);
@@ -374,7 +397,7 @@ void ExpressionVisitor::visitFunctionCall(FunctionCallAst* node)
 ///TODO: DUContext pointer?
 DUContext* ExpressionVisitor::findClassContext(IdentifierAst* className)
 {
-    DUContext* context = 0;
+    DUContext* context = nullptr;
     DeclarationPointer declaration = findDeclarationImport(ClassDeclarationType, className);
     usingDeclaration(className, declaration);
     if (declaration) {
@@ -390,7 +413,7 @@ DUContext* ExpressionVisitor::findClassContext(IdentifierAst* className)
 ///TODO: DUContext pointer?
 DUContext* ExpressionVisitor::findClassContext(NamespacedIdentifierAst* className)
 {
-    DUContext* context = 0;
+    DUContext* context = nullptr;
     const QualifiedIdentifier id = identifierForNamespace(className, m_editor);
     DeclarationPointer declaration = findDeclarationImport(ClassDeclarationType, id);
     usingDeclaration(className->namespaceNameSequence->back()->element, declaration);
@@ -438,6 +461,10 @@ void ExpressionVisitor::visitConstantOrClassConst(ConstantOrClassConstAst *node)
             //constant (created with declare('foo', 'bar')) or const Foo = 1;
             QualifiedIdentifier id = identifierForNamespace(node->constant, m_editor, true);
             DeclarationPointer declaration = findDeclarationImport(ConstantDeclarationType, id);
+            if (!declaration) {
+                id.setExplicitlyGlobal(true);
+                declaration = findDeclarationImport(ConstantDeclarationType, id);
+            }
             if (!declaration) {
                 ///TODO: is this really wanted?
                 //it could also be a global function call, without ()
@@ -567,9 +594,33 @@ void ExpressionVisitor::visitVariableProperty(VariablePropertyAst *node)
     if (node->objectProperty && node->objectProperty->objectDimList) {
         //handle $foo->bar() and $foo->baz, $foo is m_result.type()
 
-        if (m_result.type() && StructureType::Ptr::dynamicCast(m_result.type())) {
+        AbstractType::Ptr type = m_result.type();
+
+        //If the variable type is unsure, try to see if it contains a StructureType. If so, use that
+        //    (since the other types do not allow accessing properties)
+        if (type && type.cast<UnsureType>()) {
+            UnsureType::Ptr unsureType = type.cast<UnsureType>();
+            int numStructureType = 0;
+            StructureType::Ptr structureType;
+
+            for (unsigned int i = 0; i<unsureType->typesSize(); ++i) {
+                StructureType::Ptr subType = unsureType->types()[i].type<StructureType>();
+                if (subType) {
+                    structureType = subType;
+                    ++numStructureType;
+                }
+            }
+
+            //Only use the found structureType if there's exactly *one* such type
+            if (numStructureType == 1) {
+                Q_ASSERT(structureType);
+                type = AbstractType::Ptr(structureType);
+            }
+        }
+
+        if (type && StructureType::Ptr::dynamicCast(type)) {
             DUChainReadLocker lock(DUChain::lock());
-            Declaration* declaration = StructureType::Ptr::staticCast(m_result.type())->declaration(m_currentContext->topContext());
+            Declaration* declaration = StructureType::Ptr::staticCast(type)->declaration(m_currentContext->topContext());
             if (declaration) {
                 ifDebug(qCDebug(DUCHAIN) << "parent:" << declaration->toString();)
                 DUContext* context = declaration->internalContext();
@@ -597,8 +648,16 @@ void ExpressionVisitor::visitVariableProperty(VariablePropertyAst *node)
                             }
                         } else {
                             if ( !dec->isFunctionDeclaration() ) {
-                                decs << dec;
-                                ifDebug(qCDebug(DUCHAIN) << "found:" << dec->toString();)
+                                ClassMemberDeclaration *classDec = dynamic_cast<ClassMemberDeclaration*>(dec);
+                                if (classDec && classDec->accessPolicy() == Declaration::Private) {
+                                    if (declaration == dec->context()->owner()) {
+                                        decs << dec;
+                                        ifDebug(qCDebug(DUCHAIN) << "found private:" << dec->toString();)
+                                    }
+                                } else {
+                                    decs << dec;
+                                    ifDebug(qCDebug(DUCHAIN) << "found:" << dec->toString();)
+                                }
                             }
                         }
                     }
@@ -638,23 +697,16 @@ void ExpressionVisitor::visitStaticMember(StaticMemberAst* node)
 {
     //don't call DefaultVisitor::visitStaticMember(node);
     //because we would end up in visitCompoundVariableWithSimpleIndirectReference
-    if (node->variable->variable->variable) {
+    if (node->staticProperty->staticProperty->variable->variable) {
         DUContext* context = findClassContext(node->className);
         if (context) {
-            DUChainReadLocker lock(DUChain::lock());
-            m_result.setDeclarations(context->findDeclarations(identifierForNode(node->variable->variable->variable)));
-            lock.unlock();
-            if (!m_result.allDeclarations().isEmpty()) {
-                usingDeclaration(node->variable->variable->variable, m_result.allDeclarations().last());
-            } else {
-                usingDeclaration(node->variable->variable->variable, DeclarationPointer());
-            }
+            useDeclaration(node->staticProperty->staticProperty->variable, context);
         } else {
             usingDeclaration(node->className, DeclarationPointer());
             m_result.setType(AbstractType::Ptr());
         }
-        if (node->variable->offsetItemsSequence) {
-            const KDevPG::ListNode< DimListItemAst* >* it = node->variable->offsetItemsSequence->front();
+        if (node->staticProperty->offsetItemsSequence) {
+            const KDevPG::ListNode< DimListItemAst* >* it = node->staticProperty->offsetItemsSequence->front();
             do {
                 visitDimListItem(it->element);
             } while(it->hasNext() && (it = it->next));
@@ -662,6 +714,97 @@ void ExpressionVisitor::visitStaticMember(StaticMemberAst* node)
     }
 }
 
+void ExpressionVisitor::visitClassNameReference(ClassNameReferenceAst* node)
+{
+    if (node->staticProperty) {
+        DUContext* context = findClassContext(node->className->identifier);
+
+        if (context) {
+            useDeclaration(node->staticProperty->staticProperty->variable, context);
+        }
+
+        if (node->staticProperty->offsetItemsSequence) {
+            const KDevPG::ListNode< DimListItemAst* >* dim_it = node->staticProperty->offsetItemsSequence->front();
+            do {
+                visitDimListItem(dim_it->element);
+            } while(dim_it->hasNext() && (dim_it = dim_it->next));
+        }
+    }
+
+    if (node->baseVariable) {
+        DefaultVisitor::visitVariableWithoutObjects(node->baseVariable);
+    }
+
+    if (node->propertiesSequence) {
+        if (!m_result.allDeclarations().isEmpty()) {
+            DUContext* context = nullptr;
+            StructureType::Ptr type;
+
+            Declaration *declaration = nullptr;
+            const KDevPG::ListNode< ClassPropertyAst* >* it = node->propertiesSequence->front();
+
+            do {
+                // first check for property names held in variables ($object->$property)
+                if (it->element->property && it->element->property->variableWithoutObjects
+                        && it->element->property->variableWithoutObjects->variable->variable) {
+                    VariableIdentifierAst *varnode = it->element->property->variableWithoutObjects->variable->variable;
+                    useDeclaration(varnode, m_currentContext);
+                } else if (!m_result.allDeclarations().isEmpty()) {
+                    // handle array indices after normal/static properties ($object->property[$index] // $object::$property[$index])
+                    if (it->element->property && it->element->property->objectDimList && it->element->property->objectDimList->offsetItemsSequence) {
+                        const KDevPG::ListNode< DimListItemAst* >* dim_it = it->element->property->objectDimList->offsetItemsSequence->front();
+                        do {
+                            visitDimListItem(dim_it->element);
+                        } while(dim_it->hasNext() && (dim_it = dim_it->next));
+                    } else if (it->element->staticProperty && it->element->staticProperty->offsetItemsSequence) {
+                        const KDevPG::ListNode< DimListItemAst* >* dim_it = it->element->staticProperty->offsetItemsSequence->front();
+                        do {
+                            visitDimListItem(dim_it->element);
+                        } while(dim_it->hasNext() && (dim_it = dim_it->next));
+                    }
+
+                    type = m_result.allDeclarations().last()->type<StructureType>();
+
+                    if (!type) {
+                        context = nullptr;
+                        continue;
+                    }
+
+                    DUChainReadLocker lock(DUChain::lock());
+                    declaration = type->declaration(m_currentContext->topContext());
+                    lock.unlock();
+
+                    if (!declaration) {
+                        context = nullptr;
+                        continue;
+                    }
+
+                    context = declaration->internalContext();
+
+                    if (!context || context->type() != DUContext::Class) {
+                        context = nullptr;
+                        continue;
+                    }
+
+                    if (it->element->staticProperty) {
+                        // static properties ($object::$property)
+                        VariableIdentifierAst *varnode = it->element->staticProperty->staticProperty->variable;
+                        useDeclaration(varnode, context);
+                    } else if (it->element->property && it->element->property->objectDimList
+                            && it->element->property->objectDimList->variableName->name) {
+                        // normal properties ($object->property)
+                        IdentifierAst *varidnode = it->element->property->objectDimList->variableName->name;
+                        useDeclaration(varidnode, context);
+                    } else {
+                        context = nullptr;
+                    }
+                }
+            } while(it->hasNext() && (it = it->next));
+        }
+    }
+
+}
+
 void ExpressionVisitor::visitUnaryExpression(UnaryExpressionAst* node)
 {
     DefaultVisitor::visitUnaryExpression(node);
@@ -713,12 +856,44 @@ void ExpressionVisitor::visitAdditiveExpressionRest(AdditiveExpressionRestAst* n
 void ExpressionVisitor::visitRelationalExpression(RelationalExpressionAst *node)
 {
     DefaultVisitor::visitRelationalExpression(node);
-    if (node->instanceofType && node->instanceofType->identifier) {
-        const QualifiedIdentifier id = identifierForNamespace(node->instanceofType->identifier, m_editor);
+    if (node->instanceofType && node->instanceofType->className && node->instanceofType->className->identifier) {
+        const QualifiedIdentifier id = identifierForNamespace(node->instanceofType->className->identifier, m_editor);
         DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
-        usingDeclaration(node->instanceofType->identifier->namespaceNameSequence->back()->element, dec);
-        buildNamespaceUses(node->instanceofType->identifier, id);
-        m_result.setDeclaration(dec);
+        usingDeclaration(node->instanceofType->className->identifier->namespaceNameSequence->back()->element, dec);
+        buildNamespaceUses(node->instanceofType->className->identifier, id);
+
+        m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeBoolean)));
+    }
+}
+
+void ExpressionVisitor::visitRelationalExpressionRest(RelationalExpressionRestAst *node)
+{
+    DefaultVisitor::visitRelationalExpressionRest(node);
+
+    m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeBoolean)));
+}
+
+void ExpressionVisitor::visitEqualityExpressionRest(EqualityExpressionRestAst *node)
+{
+    DefaultVisitor::visitEqualityExpressionRest(node);
+
+    if (node->operation && node->operation == OperationSpaceship) {
+        m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeInt)));
+    } else {
+        m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeBoolean)));
+    }
+}
+
+void ExpressionVisitor::visitStatement(StatementAst *node)
+{
+    DefaultVisitor::visitStatement(node);
+
+    if (node->returnExpr) {
+        FunctionType::Ptr closureType = currentClosureReturnType().cast<FunctionType>();
+
+        if (closureType) {
+            closureType->setReturnType(m_result.type());
+        }
     }
 }
 
@@ -778,6 +953,30 @@ void ExpressionVisitor::buildNamespaceUses(NamespacedIdentifierAst* namespaces,
     }
 }
 
+void ExpressionVisitor::useDeclaration(VariableIdentifierAst* node, DUContext* context)
+{
+    DUChainReadLocker lock(DUChain::lock());
+    m_result.setDeclarations(context->findDeclarations(identifierForNode(node)));
+    lock.unlock();
+    if (!m_result.allDeclarations().isEmpty()) {
+        usingDeclaration(node, m_result.allDeclarations().last());
+    } else {
+        usingDeclaration(node, DeclarationPointer());
+    }
+}
+
+void ExpressionVisitor::useDeclaration(IdentifierAst* node, DUContext* context)
+{
+    DUChainReadLocker lock(DUChain::lock());
+    m_result.setDeclarations(context->findDeclarations(identifierForNode(node)));
+    lock.unlock();
+    if (!m_result.allDeclarations().isEmpty()) {
+        usingDeclaration(node, m_result.allDeclarations().last());
+    } else {
+        usingDeclaration(node, DeclarationPointer());
+    }
+}
+
 DeclarationPointer ExpressionVisitor::findDeclarationImport(DeclarationType declarationType, IdentifierAst* node)
 {
     // methods and class names are case insensitive
@@ -805,7 +1004,7 @@ Declaration* ExpressionVisitor::findVariableDeclaration(DUContext* context, Iden
                                                         CursorInRevision position, DUContext::SearchFlag flag)
 {
     QList<Declaration*> decls = context->findDeclarations(identifier, position,
-                                                        0, flag);
+                                                        nullptr, flag);
     for (int i = decls.count() - 1; i >= 0; i--) {
         Declaration *dec = decls.at(i);
         if (dec->kind() == Declaration::Instance && dynamic_cast<VariableDeclaration*>(dec)) {
@@ -813,7 +1012,7 @@ Declaration* ExpressionVisitor::findVariableDeclaration(DUContext* context, Iden
         }
     }
 
-    return NULL;
+    return nullptr;
 }
 
 }
