@@ -23,6 +23,7 @@
 #include <language/duchain/topducontext.h>
 #include <language/duchain/declaration.h>
 #include <language/duchain/types/functiontype.h>
+#include <serialization/indexedstring.h>
 
 using namespace KDevelop;
 
@@ -31,7 +32,7 @@ namespace Php {
 MagicConstantNavigationContext::MagicConstantNavigationContext(TopDUContextPointer topContext,
                                                                 const KTextEditor::Cursor& position,
                                                                 const QString& constant)
-    : AbstractNavigationContext(topContext, 0), m_position(position.line(), position.column()), m_constant(constant)
+    : AbstractNavigationContext(topContext, nullptr), m_position(position.line(), position.column()), m_constant(constant)
 {
 }
 
@@ -47,7 +48,7 @@ DUContext* findContext(TopDUContextPointer topCtx, const CursorInRevision& pos,
         ctx = ctx->parentContext();
     }
     if ( !ctx || ctx->type() != type ) {
-        return 0;
+        return nullptr;
     } else {
         return ctx;
     }
@@ -55,27 +56,35 @@ DUContext* findContext(TopDUContextPointer topCtx, const CursorInRevision& pos,
 
 QString MagicConstantNavigationContext::html(bool /*shorten*/)
 {
-    QString html = "<html><body><p><small><small>";
+    QString html = QStringLiteral("<html><body><p><small><small>");
     html += typeHighlight(i18n("magic constant"));
     html += ' ';
     html += nameHighlight(m_constant.toHtmlEscaped());
-    html += "<br/>\n";
+    html += QLatin1String("<br/>\n");
 
     QString value;
-    ///TODO: php 5.3: __DIR__, __NAMESPACE__
-    if ( m_constant == "__FILE__" ) {
-        value = m_topContext->url().str().toHtmlEscaped();
-    } else if ( m_constant == "__LINE__" ) {
+
+    if ( m_constant == QLatin1String("__FILE__") ) {
+        value = topContext()->url().str().toHtmlEscaped();
+    } else if ( m_constant == QLatin1String("__DIR__") ) {
+        value = QFileInfo(topContext()->url().str()).absoluteDir().path().toHtmlEscaped();
+    } else if ( m_constant == QLatin1String("__LINE__") ) {
         value.setNum(m_position.line + 1);
-    } else if ( m_constant == "__CLASS__" ) {
-        if ( DUContext* ctx = findContext(m_topContext, m_position, DUContext::Class) ) {
+    } else if ( m_constant == QLatin1String("__CLASS__") ) {
+        if ( DUContext* ctx = findContext(topContext(), m_position, DUContext::Class) ) {
             value = codeHighlight(ctx->localScopeIdentifier().toString().toHtmlEscaped());
         } else {
             value = commentHighlight(i18n("empty (not inside a class)"));
         }
-    } else if ( m_constant == "__METHOD__" ) {
+    } else if ( m_constant == QLatin1String("__TRAIT__") ) {
+        if ( DUContext* ctx = findContext(topContext(), m_position, DUContext::Class) ) {
+            value = codeHighlight(ctx->localScopeIdentifier().toString().toHtmlEscaped());
+        } else {
+            value = commentHighlight(i18n("empty (not inside a trait)"));
+        }
+    } else if ( m_constant == QLatin1String("__METHOD__") ) {
         CursorInRevision pos = m_position;
-        while ( DUContext* ctx = findContext(m_topContext, pos, DUContext::Other) ) {
+        while ( DUContext* ctx = findContext(topContext(), pos, DUContext::Other) ) {
             if ( !ctx->parentContext() ) {
                 break;
             }
@@ -92,9 +101,9 @@ QString MagicConstantNavigationContext::html(bool /*shorten*/)
         if ( value.isEmpty() ) {
             value = commentHighlight(i18n("empty (not inside a method)"));
         }
-    } else if ( m_constant == "__FUNCTION__" ) {
+    } else if ( m_constant == QLatin1String("__FUNCTION__") ) {
         CursorInRevision pos = m_position;
-        if ( DUContext* ctx = findContext(m_topContext, pos, DUContext::Other) ) {
+        if ( DUContext* ctx = findContext(topContext(), pos, DUContext::Other) ) {
             if ( ctx->owner() && ctx->owner()->type<FunctionType>() ) {
                 value = codeHighlight(ctx->localScopeIdentifier().toString().toHtmlEscaped());
             }
@@ -102,8 +111,8 @@ QString MagicConstantNavigationContext::html(bool /*shorten*/)
         if ( value.isEmpty() ) {
             value = commentHighlight(i18n("empty (not inside a function)"));
         }
-    } else if ( m_constant == "__NAMESPACE__" ) {
-        if ( DUContext* ctx = findContext(m_topContext, m_position, DUContext::Namespace) ) {
+    } else if ( m_constant == QLatin1String("__NAMESPACE__") ) {
+        if ( DUContext* ctx = findContext(topContext(), m_position, DUContext::Namespace) ) {
             if ( ctx->owner() && ctx->owner()->kind() == Declaration::Namespace ) {
                 value = codeHighlight(ctx->localScopeIdentifier().toString().toHtmlEscaped());
             }
@@ -115,7 +124,7 @@ QString MagicConstantNavigationContext::html(bool /*shorten*/)
 
     html += i18n("current value: %1", value);
 
-    html += "</small></small></p></body></html>";
+    html += QLatin1String("</small></small></p></body></html>");
 
     return html;
 }
