@@ -45,6 +45,8 @@ private slots:
     void testForeachImportedIdentifier();
     void testUpdateForeach();
     void testTodoExtractorReparse();
+    void testIteratorForeachReparse();
+    void testNamespacedIdentifierInPST();
 private:
     KDevelop::TestProjectController* m_projectController;
 };
