@@ -22,12 +22,15 @@
 
 #include <language/duchain/duchain.h>
 #include <language/duchain/duchainlock.h>
+#include <language/duchain/problem.h>
 
 #include "../declarations/classdeclaration.h"
 #include "../declarations/variabledeclaration.h"
 #include "../declarations/traitmethodaliasdeclaration.h"
 #include "../declarations/traitmemberaliasdeclaration.h"
 
+#include "../types/structuretype.h"
+
 using namespace KDevelop;
 
 QTEST_MAIN(Php::TestUses)
@@ -141,6 +144,22 @@ void TestUses::memberVariable()
     QCOMPARE(var->uses().keys().first(), IndexedString(QUrl("file:///internal/usestest/memberVariable.php")));
 }
 
+void TestUses::implicitMemberVariable()
+{
+
+    //                 0         1         2         3         4         5         6         7
+    //                 01234567890123456789012345678901234567890123456789012345678901234567890123456789
+    QByteArray method("<? $x = new A(); $x->y = 1; $x->y = 2; class A {}");
+    TopDUContext* top = parse(method, DumpNone);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock(DUChain::lock());
+    Declaration* var = top->childContexts().first()->localDeclarations().first();
+    QList<RangeInRevision> ranges;
+    ranges << RangeInRevision(0, 21, 0, 22) << RangeInRevision(0, 32, 0, 33);
+    compareUses(var, ranges);
+    QVERIFY(var->range() == RangeInRevision(0, 21, 0, 22));
+}
+
 void TestUses::variable()
 {
     //                        0         1         2         3         4         5         6         7
@@ -1125,7 +1144,236 @@ void TestUses::instanceof()
     compareUses(a, QList<RangeInRevision>()
                     << RangeInRevision(1, 9, 1, 10)
                     << RangeInRevision(2, 19, 2, 20));
+}
 
+void TestUses::instanceofClassProperty()
+{
+    //                         0         1         2         3         4         5
+    //                         012345678901234567890123456789012345678901234567890123456789
+    TopDUContext* top = parse("<? class A { /** @var B **/ public $foo; }\n"
+                              "class B { /** @var A **/ public $bar; }\n"
+                              "$a = new A(); $a instanceof $a->foo->bar->foo;\n", DumpNone);
+
+    QVERIFY(top);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock;
+
+    QVERIFY(top->problems().isEmpty());
+
+    QVERIFY(!top->parentContext());
+    QCOMPARE(top->childContexts().count(), 2);
+    QCOMPARE(top->localDeclarations().count(), 3);
+
+    Declaration* dec = top->localDeclarations().at(0);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 9, 2, 10));
+
+    dec = top->localDeclarations().at(2);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    StructureType::Ptr classType = dec->type<StructureType>();
+    QVERIFY(classType);
+    QCOMPARE(classType->qualifiedIdentifier(), QualifiedIdentifier("a"));
+    QVERIFY(classType->equals(dec->abstractType().data()));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 14, 2, 16)
+                    << RangeInRevision(2, 28, 2, 30));
+
+    dec = top->childContexts().at(0)->localDeclarations().at(0);
+    QVERIFY(dec);
+    QCOMPARE(dec->identifier(), Identifier("foo"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 32, 2, 35)
+                    << RangeInRevision(2, 42, 2, 45));
+
+    dec = top->childContexts().at(1)->localDeclarations().at(0);
+    QVERIFY(dec);
+    QCOMPARE(dec->identifier(), Identifier("bar"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 37, 2, 40));
+}
+
+void TestUses::instanceofStaticProperty()
+{
+    //                         0         1         2         3         4         5
+    //                         012345678901234567890123456789012345678901234567890123456789
+    TopDUContext* top = parse("<? class A { /** @var B **/ public static $foo; }\n"
+                              "class B { /** @var A **/ public static $bar; }\n"
+                              "$a = new A(); $a instanceof A::$foo::$bar::$foo;\n", DumpNone);
+
+    QVERIFY(top);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock;
+
+    QVERIFY(top->problems().isEmpty());
+
+    QVERIFY(!top->parentContext());
+    QCOMPARE(top->childContexts().count(), 2);
+    QCOMPARE(top->localDeclarations().count(), 3);
+
+    Declaration* dec = top->localDeclarations().at(0);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 9, 2, 10)
+                    << RangeInRevision(2, 28, 2, 29));
+
+    dec = top->localDeclarations().at(2);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    StructureType::Ptr classType = dec->type<StructureType>();
+    QVERIFY(classType);
+    QCOMPARE(classType->qualifiedIdentifier(), QualifiedIdentifier("a"));
+    QVERIFY(classType->equals(dec->abstractType().data()));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 14, 2, 16));
+
+    dec = top->childContexts().at(0)->localDeclarations().at(0);
+    QVERIFY(dec);
+    QCOMPARE(dec->identifier(), Identifier("foo"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 31, 2, 35)
+                    << RangeInRevision(2, 43, 2, 47));
+
+    dec = top->childContexts().at(1)->localDeclarations().at(0);
+    QVERIFY(dec);
+    QCOMPARE(dec->identifier(), Identifier("bar"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 37, 2, 41));
+}
+
+void TestUses::instanceofMixedProperty()
+{
+    //                         0         1         2         3         4         5
+    //                         012345678901234567890123456789012345678901234567890123456789
+    TopDUContext* top = parse("<? class A { /** @var B **/ public static $foo; }\n"
+                              "class B { /** @var A **/ public $bar; }\n"
+                              "$a = new A(); $a instanceof A::$foo->bar::$foo;\n", DumpNone);
+
+    QVERIFY(top);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock;
+
+    QVERIFY(top->problems().isEmpty());
+
+    QVERIFY(!top->parentContext());
+    QCOMPARE(top->childContexts().count(), 2);
+    QCOMPARE(top->localDeclarations().count(), 3);
+
+    Declaration* dec = top->localDeclarations().at(0);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 9, 2, 10)
+                    << RangeInRevision(2, 28, 2, 29));
+
+    dec = top->localDeclarations().at(2);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    StructureType::Ptr classType = dec->type<StructureType>();
+    QVERIFY(classType);
+    QCOMPARE(classType->qualifiedIdentifier(), QualifiedIdentifier("a"));
+    QVERIFY(classType->equals(dec->abstractType().data()));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 14, 2, 16));
+
+    dec = top->childContexts().at(0)->localDeclarations().at(0);
+    QVERIFY(dec);
+    QCOMPARE(dec->identifier(), Identifier("foo"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 31, 2, 35)
+                    << RangeInRevision(2, 42, 2, 46));
+
+    dec = top->childContexts().at(1)->localDeclarations().at(0);
+    QVERIFY(dec);
+    QCOMPARE(dec->identifier(), Identifier("bar"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 37, 2, 40));
+}
+
+void TestUses::instanceofVariableProperty()
+{
+    //                         0         1         2         3         4         5
+    //                         012345678901234567890123456789012345678901234567890123456789
+    TopDUContext* top = parse("<? class A { /** @var B **/ public $foo; }\n"
+                              "class B { /** @var A **/ public $bar; }\n"
+                              "$foo = 'foo'; $bar = 'bar';\n"
+                              "$a = new A(); $a instanceof $a->$foo->$bar->$foo;\n", DumpNone);
+
+    QVERIFY(top);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock;
+
+    QVERIFY(top->problems().isEmpty());
+
+    QVERIFY(!top->parentContext());
+    QCOMPARE(top->childContexts().count(), 2);
+    QCOMPARE(top->localDeclarations().count(), 5);
+
+    Declaration* dec = top->localDeclarations().at(0);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(3, 9, 3, 10));
+
+    dec = top->localDeclarations().at(4);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    StructureType::Ptr classType = dec->type<StructureType>();
+    QVERIFY(classType);
+    QCOMPARE(classType->qualifiedIdentifier(), QualifiedIdentifier("a"));
+    QVERIFY(classType->equals(dec->abstractType().data()));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(3, 14, 3, 16)
+                    << RangeInRevision(3, 28, 3, 30));
+
+    dec = top->localDeclarations().at(2);
+    QCOMPARE(dec->identifier(), Identifier("foo"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(3, 32, 3, 36)
+                    << RangeInRevision(3, 44, 3, 48));
+
+    dec = top->localDeclarations().at(3);
+    QCOMPARE(dec->identifier(), Identifier("bar"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(3, 38, 3, 42));
+}
+
+void TestUses::instanceofPropertyArrayAccess()
+{
+    //                         0         1         2         3         4         5
+    //                         012345678901234567890123456789012345678901234567890123456789
+    TopDUContext* top = parse("<? class A { /** @var array **/ public $foo; }\n"
+                              "class B { /** @var array **/ public static $bar; }\n"
+                              "$a = new A(); $a instanceof $a->foo[0]::$bar[0]->foo;\n", DumpNone);
+
+    QVERIFY(top);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock;
+
+    QVERIFY(top->problems().isEmpty());
+
+    QVERIFY(!top->parentContext());
+    QCOMPARE(top->childContexts().count(), 2);
+    QCOMPARE(top->localDeclarations().count(), 3);
+
+    Declaration* dec = top->localDeclarations().at(0);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 9, 2, 10));
+
+    dec = top->localDeclarations().at(2);
+    QCOMPARE(dec->identifier(), Identifier("a"));
+    StructureType::Ptr classType = dec->type<StructureType>();
+    QVERIFY(classType);
+    QCOMPARE(classType->qualifiedIdentifier(), QualifiedIdentifier("a"));
+    QVERIFY(classType->equals(dec->abstractType().data()));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 14, 2, 16)
+                    << RangeInRevision(2, 28, 2, 30));
+
+    dec = top->childContexts().at(0)->localDeclarations().at(0);
+    QCOMPARE(dec->identifier(), Identifier("foo"));
+    compareUses(dec, QList<RangeInRevision>()
+                    << RangeInRevision(2, 32, 2, 35));
+
+    dec = top->childContexts().at(1)->localDeclarations().at(0);
+    QCOMPARE(dec->identifier(), Identifier("bar"));
+    QVERIFY(dec->uses().isEmpty());
 }
 
 void TestUses::classNameString()
@@ -1145,8 +1393,8 @@ void TestUses::classNameString()
 
 void TestUses::useTrait()
 {
-    //                         0         1         2         3         4         5         6         7
-    //                         01234567890123456789012345678901234567890123456789012345678901234567890123456789
+    //                         0         1         2         3         4         5         6         7         8
+    //                         012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789
     TopDUContext* top = parse("<?php\n"
                               "trait A { public function one(){} public function two(){} }\n"
                               "trait B { public function one(){} private function three(){} }\n"
@@ -1154,9 +1402,11 @@ void TestUses::useTrait()
                               "class Foo { use A;}\n"
                               "class Bar { use A,B { A::one insteadof B; B::three as public four; } }\n"
                               "class Baz { use A,B,C { C::one insteadof A,B; C::five as six; } }\n"
+                              "class Boo { use A,B,C { A::one insteadof B,C; A::two as switch; B::three as public; C::five as public static; } }\n"
                               "$a = new Foo(); $a->one(); $a->two();\n"
                               "$b = new Bar(); $b->one(); $b->two(); $b->four();\n"
-                              "$c = new Baz(); $c->one(); $c->two(); $c->baz; $c::six();\n", DumpAll);
+                              "$c = new Baz(); $c->one(); $c->two(); $c->baz; $c::six();\n"
+                              "$d = new Boo(); $d->one(); $d->switch(); $d->three(); $d->static();\n", DumpAll);
     QVERIFY(top);
     DUChainReleaser releaseTop(top);
     DUChainWriteLocker lock;
@@ -1166,50 +1416,59 @@ void TestUses::useTrait()
     TraitMethodAliasDeclaration* method;
     TraitMemberAliasDeclaration* member;
 
-    QCOMPARE(topDecs.size(), 9);
+    QCOMPARE(topDecs.size(), 11);
 
     dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("a"))).first();
     compareUses(dec, QList<RangeInRevision>() << RangeInRevision(4, 16, 4, 17)
                                           << RangeInRevision(5, 16, 5, 17)
                                           << RangeInRevision(5, 22, 5, 23)
                                           << RangeInRevision(6, 16, 6, 17)
-                                          << RangeInRevision(6, 41, 6, 42) );
+                                          << RangeInRevision(6, 41, 6, 42)
+                                          << RangeInRevision(7, 16, 7, 17)
+                                          << RangeInRevision(7, 24, 7, 25)
+                                          << RangeInRevision(7, 46, 7, 47) );
 
     dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("b"))).first();
     compareUses(dec, QList<RangeInRevision>() << RangeInRevision(5, 18, 5, 19)
                                           << RangeInRevision(5, 39, 5, 40)
                                           << RangeInRevision(5, 42, 5, 43)
                                           << RangeInRevision(6, 18, 6, 19)
-                                          << RangeInRevision(6, 43, 6, 44) );
+                                          << RangeInRevision(6, 43, 6, 44)
+                                          << RangeInRevision(7, 18, 7, 19)
+                                          << RangeInRevision(7, 41, 7, 42)
+                                          << RangeInRevision(7, 64, 7, 65) );
 
     dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("c"))).first();
     compareUses(dec, QList<RangeInRevision>() << RangeInRevision(6, 20, 6, 21)
                                           << RangeInRevision(6, 24, 6, 25)
-                                          << RangeInRevision(6, 46, 6, 47) );
+                                          << RangeInRevision(6, 46, 6, 47)
+                                          << RangeInRevision(7, 20, 7, 21)
+                                          << RangeInRevision(7, 43, 7, 44)
+                                          << RangeInRevision(7, 84, 7, 85) );
 
     dec = topDecs[3]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("one"))).first();
     method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
     QVERIFY(method);
     QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("a"));
-    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(7, 20, 7, 23) );
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(8, 20, 8, 23) );
 
     dec = topDecs[3]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("two"))).first();
     method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
     QVERIFY(method);
     QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("a"));
-    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(7, 31, 7, 34) );
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(8, 31, 8, 34) );
 
     dec = topDecs[4]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("one"))).first();
     method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
     QVERIFY(method);
     QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("a"));
-    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(8, 20, 8, 23) );
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(9, 20, 9, 23) );
 
     dec = topDecs[4]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("two"))).first();
     method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
     QVERIFY(method);
     QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("a"));
-    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(8, 31, 8, 34) );
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(9, 31, 9, 34) );
 
     dec = topDecs[4]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("four"))).first();
     method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
@@ -1217,25 +1476,25 @@ void TestUses::useTrait()
     QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("b"));
     QCOMPARE(method->aliasedDeclaration().data()->identifier(), Identifier("three"));
     QCOMPARE(method->accessPolicy(), Declaration::AccessPolicy::Public);
-    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(8, 42, 8, 46) );
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(9, 42, 9, 46) );
 
     dec = topDecs[5]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("one"))).first();
     method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
     QVERIFY(method);
     QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("c"));
-    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(9, 20, 9, 23) );
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(10, 20, 10, 23) );
 
     dec = topDecs[5]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("two"))).first();
     method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
     QVERIFY(method);
     QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("a"));
-    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(9, 31, 9, 34) );
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(10, 31, 10, 34) );
 
     dec = topDecs[5]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("baz"))).first();
     member = dynamic_cast<TraitMemberAliasDeclaration*>(dec);
     QVERIFY(member);
     QCOMPARE(member->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("c"));
-    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(9, 42, 9, 45) );
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(10, 42, 10, 45) );
 
     dec = topDecs[5]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("six"))).first();
     method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
@@ -1243,7 +1502,35 @@ void TestUses::useTrait()
     QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("c"));
     QCOMPARE(method->aliasedDeclaration().data()->identifier(), Identifier("five"));
     QVERIFY(method->isStatic());
-    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(9, 51, 9, 54) );
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(10, 51, 10, 54) );
+
+    dec = topDecs[6]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("one"))).first();
+    method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
+    QVERIFY(method);
+    QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("a"));
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(11, 20, 11, 23) );
+
+    dec = topDecs[6]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("switch"))).first();
+    method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
+    QVERIFY(method);
+    QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("a"));
+    QCOMPARE(method->aliasedDeclaration().data()->identifier(), Identifier("two"));
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(11, 31, 11, 37) );
+
+    dec = topDecs[6]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("three"))).first();
+    method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
+    QVERIFY(method);
+    QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("b"));
+    QCOMPARE(method->aliasedDeclaration().data()->identifier(), Identifier("three"));
+    QCOMPARE(method->accessPolicy(), Declaration::AccessPolicy::Public);
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(11, 45, 11, 50) );
+
+    dec = topDecs[6]->internalContext()->findLocalDeclarations(Identifier(QStringLiteral("static"))).first();
+    method = dynamic_cast<TraitMethodAliasDeclaration*>(dec);
+    QVERIFY(method);
+    QCOMPARE(method->aliasedDeclaration().data()->context()->owner()->identifier(), Identifier("c"));
+    QCOMPARE(method->aliasedDeclaration().data()->identifier(), Identifier("five"));
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(11, 58, 11, 64) );
 }
 
 void TestUses::exceptionFinally()
