@@ -36,6 +36,7 @@ private slots:
     void memberFunctionCall();
     void unsureMemberFunctionCall();
     void memberVariable();
+    void implicitMemberVariable();
     void variable();
     void varInString();
     void variableInNamespace();
@@ -84,6 +85,11 @@ private slots:
     void closures();
     void closureTypehints();
     void instanceof();
+    void instanceofClassProperty();
+    void instanceofStaticProperty();
+    void instanceofMixedProperty();
+    void instanceofVariableProperty();
+    void instanceofPropertyArrayAccess();
     void classNameString();
     void useTrait();
     void exceptionFinally();
