@@ -65,7 +65,7 @@ class LanguageSupport : public KDevelop::IPlugin, public KDevelop::ILanguageSupp
 
 public:
     explicit LanguageSupport(QObject *parent, const QVariantList& args = QVariantList());
-    virtual ~LanguageSupport();
+    ~LanguageSupport() override;
     /*Name Of the Language*/
     QString name() const override;
     /*Parsejob used by background parser to parse given Url*/
@@ -78,7 +78,7 @@ public:
     /**
      * @returns the ContextMenuExtension for the Php plugin.
      */
-    KDevelop::ContextMenuExtension contextMenuExtension(KDevelop::Context *context) override;
+    KDevelop::ContextMenuExtension contextMenuExtension(KDevelop::Context* context, QWidget* parent) override;
 
     QWidget* specialLanguageObjectNavigationWidget(const QUrl& url, const KTextEditor::Cursor& position) override;
     KTextEditor::Range specialLanguageObjectRange(const QUrl& url, const KTextEditor::Cursor& position) override;
